From e4819b66649d4963a0661cba603c2967b02a6b80 Mon Sep 17 00:00:00 2001
From: Marijn van Vliet <marijn.vanvliet@med.kuleuven.be>
Date: Thu, 15 Oct 2009 15:01:55 +0000
Subject: [PATCH] * Changing styles no longer affects existing windows *
 glk_style_hint_clear() now works as intended * Support for changing
 stylehints on textgrids

There are now four sets of tags stored inside two style_sets:
default->text_grid
default->text_buffer
current->text_grid
current->text_buffer

The default styles get loaded from CSS (hardcoded if not present in CSS)
The current styles get changed all the time and copied to a new window upon
creation.


git-svn-id: http://lassie.dyndns-server.com/svn/gargoyle-gtk@144 ddfedd41-794f-dd11-ae45-00112f111e67
---
 libchimara/chimara-glk-private.h |   6 +-
 libchimara/chimara-glk.c         |   3 +-
 libchimara/strio.c               |   4 +-
 libchimara/style.c               | 373 +++++++++++++++++++++++++------
 libchimara/style.h               |  20 ++
 libchimara/window.c              |   6 +-
 tests/style.css                  |  21 +-
 7 files changed, 344 insertions(+), 89 deletions(-)

diff --git a/libchimara/chimara-glk.c b/libchimara/chimara-glk.c
index 40fddfe..6669dc6 100644
--- a/libchimara/chimara-glk.c
+++ b/libchimara/chimara-glk.c
@@ -84,7 +84,8 @@ chimara_glk_init(ChimaraGlk *self)
 	priv->default_font_desc = pango_font_description_from_string("Sans");
 	priv->monospace_font_desc = pango_font_description_from_string("Monospace");
 	priv->css_file = "style.css";
-	priv->default_styles = g_hash_table_new(g_str_hash, g_str_equal);
+	priv->default_styles = g_new0(StyleSet,1);
+	priv->current_styles = g_new0(StyleSet,1);
     priv->program = NULL;
     priv->thread = NULL;
     priv->event_queue = g_queue_new();
diff --git a/libchimara/strio.c b/libchimara/strio.c
index 3153e3e..f01099a 100644
--- a/libchimara/strio.c
+++ b/libchimara/strio.c
@@ -43,7 +43,7 @@ write_utf8_to_grid(winid_t win, gchar *s)
         GtkTextIter end = start;
         gtk_text_iter_forward_to_line_end(&end);
         gtk_text_buffer_delete(buffer, &start, &end);
-        gtk_text_buffer_insert(buffer, &start, s + (length - chars_left), available_space);
+        gtk_text_buffer_insert_with_tags_by_name(buffer, &start, s + (length - chars_left), available_space, win->window_stream->style, NULL);
         chars_left -= available_space;
         gtk_text_iter_forward_line(&start);
         available_space = win->width;
@@ -53,7 +53,7 @@ write_utf8_to_grid(winid_t win, gchar *s)
         GtkTextIter end = start;
         gtk_text_iter_forward_chars(&end, chars_left);
         gtk_text_buffer_delete(buffer, &start, &end);
-        gtk_text_buffer_insert(buffer, &start, s + (length - chars_left), -1);
+        gtk_text_buffer_insert_with_tags_by_name(buffer, &start, s + (length - chars_left), -1, win->window_stream->style, NULL);
     }
     
     gtk_text_buffer_move_mark(buffer, cursor, &start);
diff --git a/libchimara/style.c b/libchimara/style.c
index 07a10fe..f6093d9 100644
--- a/libchimara/style.c
+++ b/libchimara/style.c
@@ -4,10 +4,13 @@
 
 extern GPrivate *glk_data_key;
 static gboolean chimara_style_initialized = FALSE;
+
 static gboolean style_accept(GScanner *scanner, GTokenType token);
-static gboolean style_accept_style_identifier(GScanner *scanner);
+static gboolean style_accept_style_selector(GScanner *scanner);
 static gboolean style_accept_style_hint(GScanner *scanner, GtkTextTag *current_tag);
 static void style_add_tag_to_textbuffer(gpointer key, gpointer tag, gpointer tag_table);
+static void style_table_copy(gpointer key, gpointer tag, gpointer target_table);
+static GtkTextTag* gtk_text_tag_copy(GtkTextTag *tag);
 
 /**
  * glk_set_style:
@@ -31,25 +34,15 @@ glk_set_style(glui32 styl)
 }
 
 /* Internal function: mapping from style enum to tag name */
-static gchar *
+static gchar*
 get_tag_name(glui32 style)
 {
-	switch(style) {
-		case style_Normal: return "normal";
-		case style_Emphasized: return "emphasized";
-		case style_Preformatted: return "preformatted";
-		case style_Header: return "header";
-		case style_Subheader: return "subheader";
-		case style_Alert: return "alert";
-		case style_Note: return "note";
-		case style_BlockQuote: return "block-quote";
-		case style_Input: return "input";
-		case style_User1: return "user1";
-		case style_User2: return "user2";
+	if(style >= style_NUMSTYLES) {
+		WARNING("Unsupported style");
+		return "normal";
+	} else {
+		return (gchar*) TAG_NAMES[style];
 	}
-
-	WARNING("Unsupported style");
-	return "normal";
 }
 
 /** 
@@ -75,23 +68,93 @@ style_init_textbuffer(GtkTextBuffer *buffer)
 	}
 
 	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
-	g_hash_table_foreach(glk_data->default_styles, style_add_tag_to_textbuffer, gtk_text_buffer_get_tag_table(buffer));
+	g_hash_table_foreach(glk_data->current_styles->text_buffer, style_add_tag_to_textbuffer, gtk_text_buffer_get_tag_table(buffer));
+}
+
+/* Internal function: call this to initialize the default styles to a textgrid. */
+void
+style_init_textgrid(GtkTextBuffer *buffer)
+{
+	g_return_if_fail(buffer != NULL);
+
+	if( G_UNLIKELY(!chimara_style_initialized) ) {
+		style_init();
+	}
+
+	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
+	g_hash_table_foreach(glk_data->current_styles->text_grid, style_add_tag_to_textbuffer, gtk_text_buffer_get_tag_table(buffer));
 }
 
+/* Internal function used to iterate over the default text tag table, applying them to a textbuffer */
 static void
 style_add_tag_to_textbuffer(gpointer key, gpointer tag, gpointer tag_table)
 {
-	gtk_text_tag_table_add(tag_table, tag);
+	gtk_text_tag_table_add( tag_table, gtk_text_tag_copy(tag) );
 }
 
+/* Internal function used to iterate over a style table, copying it */
+static void
+style_table_copy(gpointer key, gpointer tag, gpointer target_table)
+{
+	g_return_if_fail(key != NULL);
+	g_return_if_fail(tag != NULL);
+	g_return_if_fail(target_table != NULL);
+
+	g_hash_table_insert(target_table, key, gtk_text_tag_copy( GTK_TEXT_TAG(tag) ));
+}
+
+/* Internal function that copies a text tag */
+static GtkTextTag*
+gtk_text_tag_copy(GtkTextTag *tag)
+{
+	GtkTextTag *copy;
+
+	g_return_val_if_fail(tag != NULL, NULL);
+
+	copy = gtk_text_tag_new(tag->name);
+	gtk_text_attributes_copy_values(tag->values, copy->values);
+
+	#define _COPY_FLAG(flag) copy->flag = tag->flag
+		_COPY_FLAG (bg_color_set);
+		_COPY_FLAG (bg_color_set);
+		_COPY_FLAG (bg_stipple_set);
+		_COPY_FLAG (fg_color_set);
+		_COPY_FLAG (fg_stipple_set);
+		_COPY_FLAG (justification_set);
+		_COPY_FLAG (left_margin_set);
+		_COPY_FLAG (indent_set);
+		_COPY_FLAG (rise_set);
+		_COPY_FLAG (strikethrough_set);
+		_COPY_FLAG (right_margin_set);
+		_COPY_FLAG (pixels_above_lines_set);
+		_COPY_FLAG (pixels_below_lines_set);
+		_COPY_FLAG (pixels_inside_wrap_set);
+		_COPY_FLAG (tabs_set);
+		_COPY_FLAG (underline_set);
+		_COPY_FLAG (wrap_mode_set);
+		_COPY_FLAG (bg_full_height_set);
+		_COPY_FLAG (invisible_set);
+		_COPY_FLAG (editable_set);
+		_COPY_FLAG (language_set);
+	#undef _COPY_FLAG
+
+	return copy;
+}
+    
+/* Internal function that reads the default styles from a CSS file */
 void
 style_init()
 {
 	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
+	GHashTable *default_text_grid_styles = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_object_unref);
+	GHashTable *default_text_buffer_styles = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_object_unref);
+	GHashTable *current_text_grid_styles = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_object_unref);
+	GHashTable *current_text_buffer_styles = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_object_unref);
 	GtkTextTag *tag;
 
 	/* Create the CSS file scanner */
 	GScanner *scanner = g_scanner_new(NULL);
+
 	int f = open(glk_data->css_file, O_RDONLY);
 	g_return_if_fail(f != -1);
 	g_scanner_input_file(scanner, f);
@@ -100,55 +163,70 @@ style_init()
 	scanner->config->cset_identifier_nth = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-_0123456789";
 	scanner->config->symbol_2_token = TRUE;
 	scanner->config->cpair_comment_single = NULL;
+	scanner->config->scan_float = FALSE;
 
 	/* Initialise the default styles */
-	g_hash_table_insert(glk_data->default_styles, "normal", gtk_text_tag_new("normal"));
+	g_hash_table_insert(default_text_grid_styles, "normal", gtk_text_tag_new("normal"));
 
 	tag = gtk_text_tag_new("emphasized");
 	g_object_set(tag, "style", PANGO_STYLE_ITALIC, "style-set", TRUE, NULL);
-	g_hash_table_insert(glk_data->default_styles, "emphasized", tag);
+	g_hash_table_insert(default_text_grid_styles, "emphasized", tag);
 
 	tag = gtk_text_tag_new("preformatted");
 	g_object_set(tag, "font-desc", glk_data->monospace_font_desc, NULL);
-	g_hash_table_insert(glk_data->default_styles, "preformatted", tag);
+	g_hash_table_insert(default_text_grid_styles, "preformatted", tag);
 
 	tag = gtk_text_tag_new("header");
 	g_object_set(tag, "size-points", 18.0, "weight", PANGO_WEIGHT_BOLD, NULL);
-	g_hash_table_insert(glk_data->default_styles, "header", tag);
+	g_hash_table_insert(default_text_grid_styles, "header", tag);
 
 	tag = gtk_text_tag_new("subheader");
 	g_object_set(tag, "size-points", 14.0, "weight", PANGO_WEIGHT_BOLD, NULL);
-	g_hash_table_insert(glk_data->default_styles, "subheader", tag);
+	g_hash_table_insert(default_text_grid_styles, "subheader", tag);
 
 	tag = gtk_text_tag_new("alert");
 	g_object_set(tag, "foreground", "#aa0000", "weight", PANGO_WEIGHT_BOLD, NULL);
-	g_hash_table_insert(glk_data->default_styles, "alert", tag);
+	g_hash_table_insert(default_text_grid_styles, "alert", tag);
 
 	tag = gtk_text_tag_new("note");
 	g_object_set(tag, "foreground", "#aaaa00", "weight", PANGO_WEIGHT_BOLD, NULL);
-	g_hash_table_insert(glk_data->default_styles, "note", tag);
+	g_hash_table_insert(default_text_grid_styles, "note", tag);
 
 	tag = gtk_text_tag_new("block-quote");
 	g_object_set(tag, "justification", GTK_JUSTIFY_CENTER, "style", PANGO_STYLE_ITALIC, NULL);
-	g_hash_table_insert(glk_data->default_styles, "block-quote", tag);
+	g_hash_table_insert(default_text_grid_styles, "block-quote", tag);
+
+	g_hash_table_insert(default_text_grid_styles, "input", gtk_text_tag_new("input"));
+	g_hash_table_insert(default_text_grid_styles, "user1", gtk_text_tag_new("user1"));
+	g_hash_table_insert(default_text_grid_styles, "user2", gtk_text_tag_new("user2"));
 
-	g_hash_table_insert(glk_data->default_styles, "input", gtk_text_tag_new("input"));
-	g_hash_table_insert(glk_data->default_styles, "user1", gtk_text_tag_new("user1"));
-	g_hash_table_insert(glk_data->default_styles, "user2", gtk_text_tag_new("user2"));
+	g_hash_table_foreach(default_text_grid_styles, style_table_copy, default_text_buffer_styles);
+	glk_data->default_styles->text_grid = default_text_grid_styles;
+	glk_data->default_styles->text_buffer = default_text_buffer_styles;
 
-	/* Run the scanner over the CSS file */
+	/* Run the scanner over the CSS file, overriding defaults */
 	while( g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) {
-		if( !style_accept_style_identifier(scanner) )
+		if( !style_accept_style_selector(scanner) )
 			break;
 	}
 
+	/* Set the current style to a copy of the default style */
+	g_hash_table_foreach(default_text_grid_styles, style_table_copy, current_text_grid_styles);
+	g_hash_table_foreach(default_text_buffer_styles, style_table_copy, current_text_buffer_styles);
+	glk_data->current_styles->text_grid = current_text_grid_styles;
+	glk_data->current_styles->text_buffer = current_text_buffer_styles;
+
 	g_scanner_destroy(scanner);
+
+	chimara_style_initialized = TRUE;
 }
 
+/* Internal function: parses a token */
 static gboolean
 style_accept(GScanner *scanner, GTokenType token)
 {
-	if( g_scanner_get_next_token(scanner) != token ) {
+	GTokenType next = g_scanner_get_next_token(scanner);
+   	if(next	!= token ) {
 		g_scanner_unexp_token(scanner, token, NULL, NULL, NULL, "CSS Error", 1);
 		return FALSE;
 	} else {
@@ -156,22 +234,42 @@ style_accept(GScanner *scanner, GTokenType token)
 	}
 }
 
+/* Internal function: parses a style selector */
 static gboolean
-style_accept_style_identifier(GScanner *scanner)
+style_accept_style_selector(GScanner *scanner)
 {
 	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
 
 	GtkTextTag *current_tag;
+	gchar *field;
 	GTokenType token = g_scanner_get_next_token(scanner);
 	GTokenValue value = g_scanner_cur_value(scanner);
 
+	if(
+		token != G_TOKEN_IDENTIFIER ||
+		( strcmp(value.v_identifier, "buffer") && strcmp(value.v_identifier, "grid") )
+	) {
+		g_scanner_error(scanner, "CSS Error: buffer/grid expected");
+		return FALSE;
+	}
+
+	field = g_strdup(value.v_identifier);
+
+	if( !style_accept(scanner, '.') )
+		return FALSE;
+
+	token = g_scanner_get_next_token(scanner);
+	value = g_scanner_cur_value(scanner);
+
 	if(token != G_TOKEN_IDENTIFIER) {
-		g_scanner_error(scanner, "CSS Error: style identifier expected");
+		g_scanner_error(scanner, "CSS Error: style selector expected");
 		return FALSE;
 	}
 
-	printf("Identifier: %s\n", value.v_identifier);
-	current_tag = g_hash_table_lookup(glk_data->default_styles, value.v_identifier);
+	if( !strcmp(field, "buffer") )
+		current_tag = g_hash_table_lookup(glk_data->default_styles->text_buffer, value.v_identifier);
+	else
+		current_tag = g_hash_table_lookup(glk_data->default_styles->text_grid, value.v_identifier);
 
 	if(current_tag == NULL) {
 		g_scanner_error(scanner, "CSS Error: invalid style identifier");
@@ -192,6 +290,7 @@ style_accept_style_identifier(GScanner *scanner)
 	return TRUE;
 }
 
+/* Internal function: parses a style hint */
 static gboolean
 style_accept_style_hint(GScanner *scanner, GtkTextTag *current_tag)
 {
@@ -205,7 +304,6 @@ style_accept_style_hint(GScanner *scanner, GtkTextTag *current_tag)
 	}
 
 	hint = g_strdup(value.v_identifier);
-	printf("Hint: %s\n", hint);
 
 	if( !style_accept(scanner, ':') )
 		return FALSE;
@@ -265,6 +363,7 @@ style_accept_style_hint(GScanner *scanner, GtkTextTag *current_tag)
 			g_scanner_error(scanner, "CSS Error: hex color expected");
 			return FALSE;
 		}
+
 		g_object_set(current_tag, "foreground", value.v_identifier, "foreground-set", TRUE, NULL);
 	}
 	else if( !strcmp(hint, "background-color") ) {
@@ -331,9 +430,12 @@ style_accept_style_hint(GScanner *scanner, GtkTextTag *current_tag)
 	return TRUE;
 }
 
+/* Internal function: parses a glk color to a #hex-value */
 static void
 color_format(glui32 val, gchar *buffer)
 {
+	g_return_if_fail(buffer != NULL);
+
 	sprintf(buffer, "#%02X%02X%02X",
 		((val & 0xff0000) >> 16),
 		((val & 0x00ff00) >> 8),
@@ -341,6 +443,14 @@ color_format(glui32 val, gchar *buffer)
 	);
 }
 
+/* Internal function: parses a GdkColor to a glk color */
+static glui32
+color_parse_gdk(GdkColor *color)
+{
+	g_return_val_if_fail(color != NULL, 0);
+	return (glui32) color->pixel;
+}
+
 /* Internal function: changes a GTK tag to correspond with the given style. */
 static void
 apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
@@ -359,12 +469,12 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 	gchar color[20];
 	switch(hint) {
 	case stylehint_Indentation:
-		g_object_set(tag_object, "left_margin", 5*val, NULL);
-		g_object_set(tag_object, "right_margin", 5*val, NULL);
+		g_object_set(tag_object, "left-margin", 5*val, "left-margin-set", TRUE, NULL);
+		g_object_set(tag_object, "right-margin", 5*val, "right-margin-set", TRUE, NULL);
 		break;
 	
 	case stylehint_ParaIndentation:
-		g_object_set(tag_object, "indent", 5*val, NULL);
+		g_object_set(tag_object, "indent", 5*val, "indent-set", TRUE, NULL);
 		break;
 
 	case stylehint_Justification:
@@ -377,7 +487,7 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 				WARNING("Unknown justification");
 				i = GTK_JUSTIFY_LEFT;
 		}
-		g_object_set(tag_object, "justification", i, NULL);
+		g_object_set(tag_object, "justification", i, "justification-set", TRUE, NULL);
 		break;
 
 	case stylehint_Weight:
@@ -387,15 +497,15 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 			case  1: i = PANGO_WEIGHT_BOLD; break;
 			default: WARNING("Unknown font weight");
 		}
-		g_object_set(tag_object, "weight", i, NULL);
+		g_object_set(tag_object, "weight", i, "weight-set", TRUE, NULL);
 		break;
 
 	case stylehint_Size:
-		g_object_set(tag_object, "size", 14+(2*val), NULL);
+		g_object_set(tag_object, "size", 14+(2*val), "size-set", TRUE, NULL);
 		break;
 
 	case stylehint_Oblique:
-		g_object_set(tag_object, "style", val ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL, NULL);
+		g_object_set(tag_object, "style", val ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL, "style-set", TRUE, NULL);
 		break;
 
 	case stylehint_Proportional:
@@ -406,9 +516,9 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 		color_format(val, color);
 
 		if(!reverse_color)
-			g_object_set(tag_object, "foreground", color, NULL);
+			g_object_set(tag_object, "foreground", color, "foreground-set", TRUE, NULL);
 		else
-			g_object_set(tag_object, "background", color, NULL);
+			g_object_set(tag_object, "background", color, "background-set", TRUE, NULL);
 
 		break;
 
@@ -416,23 +526,34 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 		color_format(val, color);
 
 		if(!reverse_color)
-			g_object_set(tag_object, "background", color, NULL);
+			g_object_set(tag_object, "background", color, "background-set", TRUE, NULL);
 		else
-			g_object_set(tag_object, "foreground", color, NULL);
+			g_object_set(tag_object, "foreground", color, "background-set", TRUE, NULL);
 
 		break;
 
 	case stylehint_ReverseColor:
 		if(reverse_color != val) {
 			/* Flip the fore- and background colors */
-			gchar* foreground_color;
-			gchar* background_color;
-			g_object_get(tag_object, "foreground", &foreground_color, NULL);
-			g_object_get(tag_object, "background", &background_color, NULL);
-			g_object_set(tag_object, "foreground", background_color, NULL);
-			g_object_set(tag_object, "background", foreground_color, NULL);
-			g_free(foreground_color);
-			g_free(background_color);
+			GdkColor* foreground_color;
+			GdkColor* background_color;
+			gint f_set, b_set = 0;
+			g_object_get(tag_object, "foreground-set", &f_set, "background-set", &b_set, NULL);
+
+			if(f_set)
+				g_object_get(tag_object, "foreground-gdk", &foreground_color, NULL);
+			if(b_set)
+				g_object_get(tag_object, "background-gdk", &background_color, NULL);
+
+			if(b_set)
+				g_object_set(tag_object, "foreground-gdk", background_color, NULL);
+			else
+				g_object_set(tag_object, "foreground", "#ffffff", NULL);
+
+			if(f_set)
+				g_object_set(tag_object, "background-gdk", foreground_color, NULL);
+			else
+				g_object_set(tag_object, "background", "#000000", NULL);
 		}
 		break;
 
@@ -440,6 +561,88 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 		WARNING("Unknown style hint");
 	}
 }
+/*Internal function: queries a text tag for the value of a given style hint */
+static gint
+query_tag(GtkTextTag *tag, glui32 hint)
+{
+	gint intval;
+	GObject *objval;
+	GdkColor *colval;
+
+	g_return_val_if_fail(tag != NULL, 0);
+
+	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
+
+	switch(hint) {
+	case stylehint_Indentation:
+		g_object_get(tag, "left_margin", &intval, NULL);
+		return intval/5;
+		break;
+	
+	case stylehint_ParaIndentation:
+		g_object_get(tag, "indent", &intval, NULL);
+		return intval/5;
+		break;
+
+	case stylehint_Justification:
+		g_object_get(tag, "justification", &intval, NULL);
+		switch(intval) {
+			case GTK_JUSTIFY_LEFT: return stylehint_just_LeftFlush; break;
+			case GTK_JUSTIFY_FILL: return stylehint_just_LeftRight; break;
+			case GTK_JUSTIFY_CENTER: return stylehint_just_Centered; break;
+			case GTK_JUSTIFY_RIGHT: return stylehint_just_RightFlush; break;
+			default: 
+				WARNING("Unknown justification");
+				return stylehint_just_LeftFlush;
+		}
+		break;
+
+	case stylehint_Weight:
+		g_object_get(tag, "weight", &intval, NULL);
+		switch(intval) {
+			case PANGO_WEIGHT_LIGHT: return -1; break;
+			case PANGO_WEIGHT_NORMAL: return 0; break;
+			case PANGO_WEIGHT_BOLD: return 1; break;
+			default: WARNING("Unknown font weight"); return 0;
+		}
+		break;
+
+	case stylehint_Size:
+		g_object_get(tag, "size", &intval, NULL);
+		return (intval/2)-14;
+		break;
+
+	case stylehint_Oblique:
+		g_object_get(tag, "style", &intval , NULL);
+		return intval == PANGO_STYLE_ITALIC ? 1 : 0;
+		break;
+
+	case stylehint_Proportional:
+		g_object_get(tag, "font-desc", &objval, NULL);
+		return objval == (GObject *)glk_data->monospace_font_desc ? 0 : 1;
+		break;
+
+	case stylehint_TextColor:
+		g_object_get(tag, "foreground-gdk", &colval, NULL);
+		return color_parse_gdk(colval);
+		break;
+
+	case stylehint_BackColor:
+		g_object_get(tag, "background-gdk", &colval, NULL);
+		return color_parse_gdk(colval);
+		break;
+
+	case stylehint_ReverseColor:
+		/* FIXME: implement this */
+		return 0;
+		break;
+
+	default:
+		WARNING("Unknown style hint");
+	}
+	
+	return 0;
+}
 
 /**
  * glk_stylehint_set:
@@ -459,28 +662,41 @@ apply_stylehint_to_tag(GtkTextTag *tag, glui32 hint, glsi32 val)
 void
 glk_stylehint_set(glui32 wintype, glui32 styl, glui32 hint, glsi32 val)
 {
-	gchar *tag_name = get_tag_name(styl);
+	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
 
-	/* Iterate over all the window and update their styles if nessecary */
-	winid_t win;
-	for(win = glk_window_iterate(NULL, NULL); win; win = glk_window_iterate(win, NULL)) {
-		if(wintype != wintype_TextBuffer)
-			continue; /* FIXME: add support for text grid windows */
+	if( G_UNLIKELY(!chimara_style_initialized) ) {
+		style_init();
+	}
 
-		if(wintype == wintype_AllTypes || glk_window_get_type(win) == wintype) {
-			GtkWidget *textview = win->widget;
-			GtkTextBuffer *textbuffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW(textview) );
-			GtkTextTagTable *table = gtk_text_buffer_get_tag_table(textbuffer);
-			GtkTextTag *to_change = gtk_text_tag_table_lookup(table, tag_name);
+	GtkTextTag *to_change;
+	if(wintype == wintype_TextBuffer || wintype == wintype_AllTypes) {
+		to_change = g_hash_table_lookup( glk_data->current_styles->text_buffer, get_tag_name(styl) );
+		apply_stylehint_to_tag(to_change, hint, val);
+	}
 
-			apply_stylehint_to_tag(to_change, hint, val);
-		}
+	if(wintype == wintype_TextGrid || wintype == wintype_AllTypes) {
+		to_change = g_hash_table_lookup( glk_data->current_styles->text_grid, get_tag_name(styl) );
+		apply_stylehint_to_tag(to_change, hint, val);
 	}
 }
 
 void
 glk_stylehint_clear(glui32 wintype, glui32 styl, glui32 hint)
 {
+	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
+	GtkTextTag *tag;
+
+	switch(wintype) {
+	case wintype_TextBuffer:
+		tag = g_hash_table_lookup( glk_data->default_styles->text_buffer, get_tag_name(styl) );
+		glk_stylehint_set( wintype, styl, hint, query_tag(tag, hint) );
+		break;
+	case wintype_TextGrid:
+		tag = g_hash_table_lookup( glk_data->default_styles->text_grid, get_tag_name(styl) );
+		glk_stylehint_set( wintype, styl, hint, query_tag(tag, hint) );
+	default:
+		return;
+	}
 }
 
 glui32
@@ -492,5 +708,20 @@ glk_style_distinguish(winid_t win, glui32 styl1, glui32 styl2)
 glui32
 glk_style_measure(winid_t win, glui32 styl, glui32 hint, glui32 *result)
 {
-	return FALSE;
+	ChimaraGlkPrivate *glk_data = g_private_get(glk_data_key);
+	GtkTextTag *tag;
+
+	switch(win->type) {
+	case wintype_TextBuffer:
+		tag = g_hash_table_lookup( glk_data->current_styles->text_buffer, get_tag_name(styl) );
+		*result = query_tag(tag, hint);
+		break;
+	case wintype_TextGrid:
+		tag = g_hash_table_lookup( glk_data->current_styles->text_grid, get_tag_name(styl) );
+		*result = query_tag(tag, hint);
+	default:
+		return FALSE;
+	}
+
+	return TRUE;
 }
diff --git a/libchimara/style.h b/libchimara/style.h
index 73cbf8c..4f67973 100644
--- a/libchimara/style.h
+++ b/libchimara/style.h
@@ -11,6 +11,26 @@
 #include "stream.h"
 
 G_GNUC_INTERNAL void style_init_textbuffer(GtkTextBuffer *buffer);
+G_GNUC_INTERNAL void style_init_textgrid(GtkTextBuffer *buffer);
 G_GNUC_INTERNAL void style_init();
 
+static const gchar* TAG_NAMES[] = {
+	"normal",
+	"emphasized",
+	"preformatted",
+	"header",
+	"subheader",
+	"alert",
+	"note",
+	"block-quote",
+	"input",
+	"user1",
+	"user2"
+};
+
+typedef struct StyleSet {
+	GHashTable *text_grid;
+	GHashTable *text_buffer;
+} StyleSet;
+
 #endif
diff --git a/libchimara/window.c b/libchimara/window.c
index 7dd14c7..c835ccb 100644
--- a/libchimara/window.c
+++ b/libchimara/window.c
@@ -458,6 +458,7 @@ glk_window_open(winid_t split, glui32 method, glui32 size, glui32 wintype,
 		case wintype_TextGrid:
 		{
 		    GtkWidget *textview = gtk_text_view_new();
+			GtkTextBuffer *textbuffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW(textview) );
 
 		    gtk_text_view_set_wrap_mode( GTK_TEXT_VIEW(textview), GTK_WRAP_NONE );
 		    gtk_text_view_set_editable( GTK_TEXT_VIEW(textview), FALSE );
@@ -479,6 +480,9 @@ glk_window_open(winid_t split, glui32 method, glui32 size, glui32 wintype,
 			/* Connect signal handlers */
 			win->keypress_handler = g_signal_connect( G_OBJECT(textview), "key-press-event", G_CALLBACK(on_window_key_press_event), win );
 			g_signal_handler_block( G_OBJECT(textview), win->keypress_handler );
+
+			/* Create the styles available to the window stream */
+			style_init_textgrid(textbuffer);
 		}
 		    break;
 		
@@ -522,7 +526,7 @@ glk_window_open(winid_t split, glui32 method, glui32 size, glui32 wintype,
 			(for line input) */
 			gtk_text_buffer_create_tag(textbuffer, "uneditable", "editable", FALSE, "editable-set", TRUE, NULL);
 
-			/* Create the default styles available to the window stream */
+			/* Create the styles available to the window stream */
 			style_init_textbuffer(textbuffer);
 
 			/* Mark the position where the user will input text */
diff --git a/tests/style.css b/tests/style.css
index 0b97e76..5c6c147 100644
--- a/tests/style.css
+++ b/tests/style.css
@@ -20,44 +20,41 @@
  * background-color (#hex-value)
  * text-align (left/right/center)
  */
-
-
-normal {
-	font-family: 'times';
+buffer.normal {
 	font-size: 12;
 }
 
-header {
+buffer.header {
 	font-size: 18;
 	font-weight: bold;
 	text-align: center;
 }
 
-subheader {
+buffer.subheader {
 	font-size: 14;
 	font-weight: bold;
 }
 
-alert {
+buffer.alert {
 	color: #aa0000;
 	font-weight: bold;
 }
 
-note {
+buffer.note {
 	color: #aaaa00;
 	font-weight: bold;
 }
 
-block-quote {
+buffer.block-quote {
 	text-align: center;
 	font-style: italic;
 }
 
-input {
+buffer.input {
 }
 
-user1 {
+buffer.user1 {
 }
 
-user2 {
+buffer.user2 {
 }
diff --git a/libchimara/chimara-glk-private.h b/libchimara/chimara-glk-private.h
index 4a274fb..5ea1e54 100644
--- a/libchimara/chimara-glk-private.h
+++ b/libchimara/chimara-glk-private.h
@@ -5,6 +5,7 @@
 #include <gmodule.h>
 #include <pango/pango.h>
 #include "glk.h"
+#include "style.h"
 #include "gi_blorb.h"
 #include "gi_dispa.h"
 #include "chimara-glk.h"
@@ -30,8 +31,9 @@ struct _ChimaraGlkPrivate {
 	guint spacing;
 	/* The CSS file to read style defaults from */
 	gchar *css_file;
-	/* Hashtable containing the default styles */
-	GHashTable *default_styles;
+	/* Hashtable containing the default and current style */
+	struct StyleSet *default_styles;
+	struct StyleSet *current_styles;
 
 	/* *** Threading data *** */
     /* Glk program loaded in widget */
-- 
2.30.2